"""
Satellite PostgreSQL database queries
=====================================

This module contains the following parsers:

SatelliteAdminSettings - command ``psql -d foreman -c 'select name, value, "default" from settings where name in (\'destroy_vm_on_host_delete\', \'unregister_delete_host\') --csv'``
-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
SatelliteComputeResources - command ``psql -d foreman -c 'select name, type from compute_resources' --csv``
-----------------------------------------------------------------------------------------------------------
SatelliteSCAStatus - command ``psql -d candlepin -c "select displayname, content_access_mode from cp_owner" --csv``
-------------------------------------------------------------------------------------------------------------------
"""

import os
import yaml
from csv import DictReader

from insights import parser, CommandParser
from insights.specs import Specs
from insights.parsers import SkipException, ParseException
from insights.parsers import keyword_search


class SatellitePostgreSQLQuery(CommandParser, list):
    """
    Parent class of satellite postgresql table queries.
    It saves the rows data into a list. Each row is saved into a dict.
    The key is the column name, the value is the value of the column.

    Resultant data structure::

        [
            {
                'name': 'abc',
                'url': '',
                'value': 'test'
            },
            {
                'name': 'def',
                'url': 'http://xx.com',
                'value': ''
            }
        ]

    Sample Output::

        name,url,value
        abc,,test
        def,http://xx.com,


    Examples:
        >>> type(query)
        <class 'insights.parsers.satellite_postgresql_query.SatellitePostgreSQLQuery'>
        >>> rows = query.search(name='abc')
        >>> len(rows)
        1
        >>> rows[0]['value']
        'test'
        >>> columns=query.get_columns()
        >>> 'url' in columns
        True
        >>> 'name' in columns
        True

    Raises:
        SkipException: when there isn't data in the table
        ParseException: when the output isn't in good csv format
    """

    def parse_content(self, content):
        if not content or len(content) == 1:
            raise SkipException("There is no data in the table")
        try:
            # keep the line break for yaml parse in some table
            reader = DictReader(os.linesep.join(content).splitlines(True))
        except Exception:
            raise ParseException("The content isn't in csv format")
        for row in reader:
            self.append(row)

    def get_columns(self):
        return list(self[0].keys())

    def search(self, **kwargs):
        """
        Get the rows by searching the table with kwargs.
        This uses the :py:func:`insights.parsers.keyword_search` function for
        searching; see its documentation for usage details. If no search
        parameters are given, no rows are returned.

        It simplify the value of the column according to actual usage.

        Returns:
            list: A list of dictionaries of rows that match the given
            search criteria.

        Examples:
            >>> query.search(name__startswith='abc') == [
            ... {'name': 'abc', 'url': '', 'value': 'test'},
            ... {'name': 'abcdef', 'url': '', 'value': 'test2'}
            ... ]
            True
            >>> query.search(name__startswith='abc', value='test') == [
            ... {'name': 'abc', 'url': '', 'value': 'test'}
            ... ]
            True
        """

        return keyword_search(self, **kwargs)


@parser(Specs.satellite_settings)
class SatelliteAdminSettings(SatellitePostgreSQLQuery):
    """
    Parse the output of the command ``psql -d foreman -c '"select name, value, "default" from settings where name in ('destroy_vm_on_host_delete', 'unregister_delete_host') --csv"``.

    Sample output::

        name,value,default
        unregister_delete_host,"--- true
        ...","--- false
        ..."
        destroy_vm_on_host_delete,,"--- true
        ..."

    Examples:
        >>> type(table)
        <class 'insights.parsers.satellite_postgresql_query.SatelliteAdminSettings'>
        >>> table.get_setting('unregister_delete_host')
        True
        >>> table.get_setting('destroy_vm_on_host_delete')
        True
    """

    def _parse_yaml(self, value):
        if value:
            try:
                return yaml.safe_load(value)
            except Exception:
                raise ParseException("Bad format value: %s" % value)
        return value

    def parse_content(self, content):
        """
        The "default" and "value" columns must be selected, or else the
        settings value can't be determined.
        The "default" and "value" column are in yaml format, it is transfer to
        python object.

        Raises:
            SkipException: when value or default column isn't found in the
                            table.
            ParseException: when the value or default in bad yaml format.
        """
        super(SatelliteAdminSettings, self).parse_content(content)
        if not all(item in self.get_columns() for item in ['default', 'value']):
            raise SkipException('No default, value columns in the table.')
        for row in self:
            row['default'] = self._parse_yaml(row['default'])
            row['value'] = self._parse_yaml(row['value'])

    def get_setting(self, setting_name):
        """
        Get the actual value of setting_name.
        If the value column isn't empty, the value of the setting_name is the
        value column, or else it's the default column.

        Args:
            setting_name (str): the value of name column which is searched in the table.

        Returns:
            It depends on the setting, maybe boolean, string, int or a list.
            None if the setting_name doesn't exist in the table.
        """
        rows = self.search(name=setting_name)
        if rows:
            value = rows[0].get('value')
            return rows[0].get('default') if value == '' else value


@parser(Specs.satellite_compute_resources)
class SatelliteComputeResources(SatellitePostgreSQLQuery):
    """
    Parse the output of the command ``psql -d foreman -c 'select name, type from compute_resources' --csv``.

    .. note::
        Please refer to its super-class :class:`insights.parsers.satellite_postgresql_query.SatellitePostgreSQLQuery` for more
        details.

    Sample output::

        name,type
        test_compute_resource1,Foreman::Model::Libvirt
        test_compute_resource2,Foreman::Model::RHV

    Examples:
        >>> type(resources_table)
        <class 'insights.parsers.satellite_postgresql_query.SatelliteComputeResources'>
        >>> rows=resources_table.search(type='Foreman::Model::Libvirt')
        >>> len(rows)
        1
        >>> rows[0]['name']
        'test_compute_resource1'
    """
    pass


@parser(Specs.satellite_sca_status)
class SatelliteSCAStatus(SatellitePostgreSQLQuery):
    """
    Parse the output of the command ``psql -d candlepin -c "select displayname, content_access_mode from cp_owner" --csv``.

    .. note::
        Please refer to its super-class :class:`insights.parsers.satellite_postgresql_query.SatellitePostgreSQLQuery` for more
        details.

    Sample output::

        displayname,content_access_mode
        Default Organization,entitlement
        Orgq,org_environment

    Examples:
        >>> type(sat_sca_info)
        <class 'insights.parsers.satellite_postgresql_query.SatelliteSCAStatus'>
        >>> sat_sca_info.sca_enabled
        True
    """

    @property
    def sca_enabled(self):
        """
        If the value of content_access_mode is "org_environment", it means the SCA is enabled for this organization.
        Return True if any organization has SCA enabled on the satellite else False
        """
        return bool(len(self.search(content_access_mode='org_environment')))



"""
KSMState - file ``/sys/kernel/mm/ksm/run``
==========================================

Parser to get the kernel samepage merging state by reading the file
``/sys/kernel/mm/ksm/run``.
"""

from insights import parser, Parser
from insights.parsers import SkipException, ParseException
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.ksmstate)
def is_running(context):
    """
    .. warning::
        This function parser is deprecated, please use :py:class:`KSMState`
        instead.

    Check if Kernel Samepage Merging is enabled. 'True' if KSM is
    on (i.e. ``/sys/kernel/mm/ksm/run`` is '1') or 'False' if not.
    """
    deprecated(is_running, "Use the `KSMState` class instead.")
    ksminfo = {}
    ksminfo['running'] = (context.content[0].split()[0] == '1')
    return ksminfo


@parser(Specs.ksmstate)
class KSMState(Parser):
    """
    Parser to get the kernel samepage merging state by reading the file
    ``/sys/kernel/mm/ksm/run``.

    Typical output of ``/sys/kernel/mm/ksm/run`` likes:

        0

    From https://www.kernel.org/doc/Documentation/vm/ksm.txt::

        set 0 to stop ksmd from running but keep merged pages,
        set 1 to run ksmd e.g. "echo 1 > /sys/kernel/mm/ksm/run",
        set 2 to stop ksmd and unmerge all pages currently merged, but leave
              mergeable areas registered for next run

    Examples:
        >>> ksm.value == '0'
        True
        >>> ksm.is_running
        False

    Raises:
        SkipException: when input is empty.
        ParseException: when the content is not expected.
    """
    def parse_content(self, content):
        if not content:
            raise SkipException
        if len(content) != 1 or content[0] not in ('0', '1', '2'):
            raise ParseException("Incorrect content: '{0}'".format(content))

        self.value = content[0]
        self.is_running = self.value == '1'



import uuid
from insights.tests import context_wrap
from insights.util.canonical_facts import (
    _filter_falsy, _safe_parse, IPs, valid_ipv4_address_or_None, valid_mac_addresses,
    valid_uuid_or_None)


def test_identity():
    assert {"foo": "bar"} == _filter_falsy({"foo": "bar"})


def test_drops_none():
    assert {"foo": "bar"} == _filter_falsy({"foo": "bar", "baz": None})


def test_drops_empty_list():
    assert {"foo": "bar"} == _filter_falsy({"foo": "bar", "baz": []})


UUID_VALID = str(uuid.uuid4())
UUID_INVALID = 'invalid uuid'


def test_valid_uuid_or_None():
    result = valid_uuid_or_None(UUID_VALID)
    assert result == UUID_VALID
    result = valid_uuid_or_None(UUID_INVALID)
    assert result is None


IPV4_VALID = [
    '192.168.1.1',
    '1.1.1.1',
    '10.110.1.254',
    '0.0.0.0',
]
IPV4_INVALID = [
    '4001:631:51:a066:250:56ff:1ea7:1696',
    '0.0',
    '192.168',
    'abc',
]


def test_valid_ipv4_address_or_None():
    for ip in IPV4_VALID:
        result = valid_ipv4_address_or_None(ip)
        assert result == ip, 'Failed valid IP address {}'.format(ip)
    for ip in IPV4_INVALID:
        result = valid_ipv4_address_or_None(ip)
        assert result is None, 'Failed invalid IP address {}'.format(ip)


class HasContent():
    def __init__(self, content=None):
        self.content = [content, ] if content is not None else []


MACS_VALID = [
    HasContent('aa:26:28:12:22:1a'),
    HasContent('   aa:26:28:12:22:1a   '),
    HasContent('aa:26:28:12:22:2a'),
    HasContent('aa:26:28:12:22:3a'),
    HasContent('00:00:00:00:00:00'),
]
MACS_INVALID = [
    HasContent('aa:26:28:12:22:1a:aa:26:28:12:22:1a'),
    HasContent('# aa:26:28:12:22:1a'),
    HasContent('00:00:00:00'),
]


def test_valid_mac_addresses():
    result = valid_mac_addresses(MACS_VALID)
    assert result == [m.content[0].strip() for m in MACS_VALID]
    result = valid_mac_addresses(MACS_INVALID)
    assert result == []


HOSTNAME_I_VALID = '192.168.1.71 10.88.0.1 172.17.0.1 172.18.0.1 10.10.121.131\n'
HOSTNAME_I_INVALID = '19f.168.1.71 0.0 2307:f1c0:ff13:f036:c0::214 f001:1a98:380:4::1d 2f07:a160:ff01:2092:e32:a5ff:ff37:7114'


def test_IPs():
    ips = IPs(context_wrap(HOSTNAME_I_VALID))
    assert ips.data == HOSTNAME_I_VALID.strip().split()
    ips = IPs(context_wrap(HOSTNAME_I_INVALID))
    assert ips.data == []


def test_safe_parse():
    result = _safe_parse(HasContent('some content'))
    assert result == 'some content'
    result = _safe_parse(HasContent())
    assert result is None
    result = _safe_parse(None)
    assert result is None
